"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

import numpy as np
from OpenGL.GL import *


def get_version():
    """
    Get the version of q3dviewer package.
    """
    try:
        from importlib.metadata import version
        return version('q3dviewer')
    except Exception:
        # Fallback if package is not installed
        return 'unknown'


def rainbow(scalars, scalar_min=0, scalar_max=255):
    range = scalar_max - scalar_min
    values = 1.0 - (scalars - scalar_min) / range
    # values = (scalars - scalar_min) / range  # using inverted color
    colors = np.zeros([scalars.shape[0], 3], dtype=np.float32)
    values = np.clip(values, 0, 1)

    h = values * 5.0 + 1.0
    i = np.floor(h).astype(int)
    f = h - i
    f[np.logical_not(i % 2)] = 1 - f[np.logical_not(i % 2)]
    n = 1 - f

    # idx = i <= 1
    colors[i <= 1, 0] = n[i <= 1] * 255
    colors[i <= 1, 1] = 0
    colors[i <= 1, 2] = 255

    colors[i == 2, 0] = 0
    colors[i == 2, 1] = n[i == 2] * 255
    colors[i == 2, 2] = 255

    colors[i == 3, 0] = 0
    colors[i == 3, 1] = 255
    colors[i == 3, 2] = n[i == 3] * 255

    colors[i == 4, 0] = n[i == 4] * 255
    colors[i == 4, 1] = 255
    colors[i == 4, 2] = 0

    colors[i >= 5, 0] = 255
    colors[i >= 5, 1] = n[i >= 5] * 255
    colors[i >= 5, 2] = 0
    return colors


def text_to_rgba(color_text, flat=False):
    """
    Convert a color text to an RGBA tuple.
    
    :param color_text: e.g. '#FF0000', '#FF0000FF', 'red', 'green', 'blue', 'yellow', 
                       'black', 'white', 'magenta', 'cyan', 'r', 'g', 'b', 'y', 'k', 'w', 'm', 'c'
    :return: RGBA tuple, e.g. (1.0, 0.0, 0.0, 1.0)
    """
    from matplotlib.colors import to_rgba

    rgba = to_rgba(color_text)
    if flat:
        r, g, b, _ = (np.array(rgba)*255).astype(np.uint32)
        falt_rgb = ((r << 16) & 0xFF0000) | \
                   ((g << 8) & 0x00FF00) | \
                   ((b << 0) & 0x0000FF)
        return falt_rgb
    else:
        return rgba
    

def set_uniform(shader, content, name):
    location = glGetUniformLocation(shader, name)
    if location == -1:
        raise ValueError(
            f"Uniform '{name}' not found in shader program {shader}.")

    if isinstance(content, int):
        glUniform1i(location, content)
    elif isinstance(content, float):
        glUniform1f(location, content)
    elif isinstance(content, np.ndarray):
        if content.ndim == 1:
            if content.shape[0] == 2:
                glUniform2f(location, *content)
            elif content.shape[0] == 3:
                glUniform3f(location, *content)
            else:
                raise ValueError(
                    f"Unsupported 1D array size: {content.shape}.")
        elif content.ndim == 2:
            if content.shape == (4, 4):
                glUniformMatrix4fv(location, 1, GL_FALSE,
                                   content.T.astype(np.float32))
            else:
                raise ValueError(
                    f"Unsupported 2D array size: {content.shape}.")
        else:
            raise ValueError(f"Unsupported array dimension: {content.ndim}.")
    else:
        raise TypeError(
            f"Unsupported type for uniform '{name}': {type(content)}.")



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from q3dviewer.Qt import QtCore
from q3dviewer.Qt.QtWidgets import QWidget, QComboBox, QVBoxLayout, QLabel, QLineEdit, QCheckBox, QGroupBox
from q3dviewer.Qt.QtGui import QKeyEvent
from q3dviewer.base_glwidget import BaseGLWidget
from q3dviewer.utils import text_to_rgba


class SettingWindow(QWidget):
    def __init__(self):
        super().__init__()
        self.combo_items = QComboBox()
        self.combo_items.currentIndexChanged.connect(self.on_combo_selection)
        main_layout = QVBoxLayout()
        main_layout.setAlignment(QtCore.Qt.AlignTop)
        main_layout.addWidget(self.combo_items)
        self.layout = QVBoxLayout()
        main_layout.addLayout(self.layout)
        self.setLayout(main_layout)
        self.setWindowTitle("Setting Window")
        self.setGeometry(200, 200, 300, 200)
        self.items = {}

    def add_setting(self, name, item):
        self.items.update({name: item})
        self.combo_items.addItem("%s(%s)" % (name, item.__class__.__name__))

    def clear_setting(self):
        while self.layout.count():
            child = self.layout.takeAt(0)
            if child.widget():
                child.widget().deleteLater()

    def on_combo_selection(self, index):
        # remove all setting of previous widget
        self.clear_setting()
        key = list(self.items.keys())
        item = self.items[key[index]]
        group_box = QGroupBox()
        group_layout = QVBoxLayout()
        item.add_setting(group_layout)
        group_box.setLayout(group_layout)
        self.layout.addWidget(group_box)


class GLWidget(BaseGLWidget):
    def __init__(self):
        self.followed_name = 'none'
        self.named_items = {}
        self.color_str = 'black'
        self.followable_item_name = None
        self.setting_window = SettingWindow()
        self.enable_show_center = True
        self.old_center = None
        super(GLWidget, self).__init__()

    def keyPressEvent(self, ev: QKeyEvent):
        if ev.key() == QtCore.Qt.Key_M:  # setting menu
            print("Open setting windows")
            self.open_setting_window()            
        else:
            super().keyPressEvent(ev)
        if ev.key() == QtCore.Qt.Key_F:  # reset follow
            if self.followable_item_name is None:
                self.initial_followable()

            if self.followed_name != 'none':
                self.followed_name = 'none'
                print("Reset follow.")
            elif len(self.followable_item_name) > 1:
                self.followed_name = self.followable_item_name[1]
                print("Set follow to ", self.followed_name)
            else:
                pass # do nothing

    def on_followable_selection(self, index):
        self.followed_name = self.followable_item_name[index]

    def mouseDoubleClickEvent(self, event):
        """Double click to set center."""
        p = self.get_point(event.x(), event.y())
        if p is not None:
            self.set_center(p)
        super().mouseDoubleClickEvent(event)

    def update(self):
        if self.followed_name != 'none':
            new_center = self.named_items[self.followed_name].T[:3, 3]
            if self.old_center is None:
                self.old_center = self.center
                return
            delta = new_center - self.old_center
            self.set_center(self.center + delta)
            self.old_center = new_center
        super().update()

    def add_setting(self, layout):
        label_color = QLabel("Set background color:")
        layout.addWidget(label_color)
        color_edit = QLineEdit()
        color_edit.setToolTip("'using hex color, i.e. #FF4500")
        color_edit.setText(self.color_str)
        color_edit.textChanged.connect(self.set_bg_color)
        layout.addWidget(color_edit)
        
        label_focus = QLabel("Set Focus:")
        combo_focus = QComboBox()
    
        if self.followable_item_name is None:
            self.initial_followable()

        for name in self.followable_item_name:
            combo_focus.addItem(name)
        combo_focus.currentIndexChanged.connect(self.on_followable_selection)
        layout.addWidget(label_focus)
        layout.addWidget(combo_focus)

        checkbox_show_center = QCheckBox("Show Center Point")
        checkbox_show_center.setChecked(self.enable_show_center)
        checkbox_show_center.stateChanged.connect(self.change_show_center)
        layout.addWidget(checkbox_show_center)

    def initial_followable(self):
        self.followable_item_name = ['none']
        for name, item in self.named_items.items():
            if item.__class__.__name__ == 'AxisItem' and not item._disable_setting:
                self.followable_item_name.append(name)

    def set_bg_color(self, color):
        try:
            self.color_str = color
            red, green, blue, alpha = text_to_rgba(color)
            self.set_color([red, green, blue, alpha])
        except ValueError:
            print("Invalid color format. Use mathplotlib color format.")

    def add_item_with_name(self, name, item):
        self.named_items.update({name: item})
        if not item._disable_setting:
            self.setting_window.add_setting(name, item)
        super().add_item(item)

    def open_setting_window(self):
        if self.setting_window.isVisible():
            self.setting_window.raise_()

        else:
            self.setting_window.show()

    def change_show_center(self, state):
        self.enable_show_center = state

    def get_camera_pose(self):
        """Get current camera pose parameters"""
        camera_pose = {
            'center': self.center.tolist() if hasattr(self.center, 'tolist') else list(self.center),
            'euler': self.euler.tolist() if hasattr(self.euler, 'tolist') else list(self.euler),
            'distance': float(self.dist),
        }
        return camera_pose

    def set_camera_pose(self, config):
        """Set camera pose from parameters"""
        if 'center' in config and 'euler' in config and 'distance' in config:
            self.set_center(config['center'])
            self.set_euler(config['euler'])
            self.set_dist(config['distance'])
        else:
            print("Invalid camera pose config")


#!/usr/bin/env python3

"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

import numpy as np
from q3dviewer.base_item import BaseItem
from OpenGL.GL import *
from OpenGL.GL import shaders
from q3dviewer.Qt.QtWidgets import QLabel, QCheckBox, QDoubleSpinBox, QSlider, QHBoxLayout, QLineEdit

import os
from q3dviewer.utils import set_uniform, text_to_rgba


class StaticMeshItem(BaseItem):
    """
    A OpenGL mesh item for rendering static 3D triangular meshes.
    Optimized for static geometry with triangle-only rendering.
    Data format: Nx9 numpy array (3 vertices per triangle, no good flag needed)
    
    Attributes:
        color (str or tuple): Accepts any valid matplotlib color (e.g., 'red', '#FF4500', (1.0, 0.5, 0.0)).
        wireframe (bool): If True, renders the mesh in wireframe mode.
    """
    def __init__(self, color='lightblue', wireframe=False):
        super(StaticMeshItem, self).__init__()
        self.wireframe = wireframe
        self.color = color
        self.flat_rgb = text_to_rgba(color, flat=True)
        
        # Static mesh buffer: N x 9 numpy array
        # Each row: [v0.x, v0.y, v0.z, v1.x, v1.y, v1.z, v2.x, v2.y, v2.z]
        self.vertices = None
        self.num_triangles = 0
        
        # OpenGL objects
        self.vao = None
        self.vbo = None
        self.program = None
        
        # Fixed rendering parameters
        self.enable_lighting = True
        self.line_width = 1.0
        self.light_color = [1.0, 1.0, 1.0]
        self.ambient_strength = 0.1
        self.diffuse_strength = 1.2
        self.specular_strength = 0.1
        self.shininess = 32.0
        self.alpha = 1.0
        
        # Settings flag
        self.need_update_setting = True
        self.need_update_buffer = True
        self.path = os.path.dirname(__file__)
    
    def add_setting(self, layout):
        """Add UI controls for mesh visualization"""
        # Wireframe toggle
        self.wireframe_box = QCheckBox("Wireframe Mode")
        self.wireframe_box.setChecked(self.wireframe)
        self.wireframe_box.toggled.connect(self.update_wireframe)
        layout.addWidget(self.wireframe_box)

        # Enable lighting toggle
        self.lighting_box = QCheckBox("Enable Lighting")
        self.lighting_box.setChecked(self.enable_lighting)
        self.lighting_box.toggled.connect(self.update_enable_lighting)
        layout.addWidget(self.lighting_box)

        # Color setting
        label_rgb = QLabel("Color:")
        label_rgb.setToolTip("Use hex color, i.e. #FF4500, or named color, i.e. 'red'")
        layout.addWidget(label_rgb)
        self.edit_rgb = QLineEdit()
        self.edit_rgb.setToolTip("Use hex color, i.e. #FF4500, or named color, i.e. 'red'")
        self.edit_rgb.setText(self.color)
        self.edit_rgb.textChanged.connect(self._on_color)
        layout.addWidget(self.edit_rgb)

        # Material property controls for Phong lighting
        if self.enable_lighting:
            # Ambient strength control
            ambient_layout = QHBoxLayout()
            ambient_label = QLabel("Ambient Strength:")
            ambient_layout.addWidget(ambient_label)
            self.ambient_slider = QSlider()
            self.ambient_slider.setOrientation(1)  # Qt.Horizontal
            self.ambient_slider.setRange(0, 100)
            self.ambient_slider.setValue(int(self.ambient_strength * 100))
            self.ambient_slider.valueChanged.connect(lambda v: self.update_ambient_strength(v / 100.0))
            ambient_layout.addWidget(self.ambient_slider)
            layout.addLayout(ambient_layout)

            # Diffuse strength control
            diffuse_layout = QHBoxLayout()
            diffuse_label = QLabel("Diffuse Strength:")
            diffuse_layout.addWidget(diffuse_label)
            self.diffuse_slider = QSlider()
            self.diffuse_slider.setOrientation(1)
            self.diffuse_slider.setRange(0, 200)
            self.diffuse_slider.setValue(int(self.diffuse_strength * 100))
            self.diffuse_slider.valueChanged.connect(lambda v: self.update_diffuse_strength(v / 100.0))
            diffuse_layout.addWidget(self.diffuse_slider)
            layout.addLayout(diffuse_layout)

            # Specular strength control
            specular_layout = QHBoxLayout()
            specular_label = QLabel("Specular Strength:")
            specular_layout.addWidget(specular_label)
            self.specular_slider = QSlider()
            self.specular_slider.setOrientation(1)
            self.specular_slider.setRange(0, 100)
            self.specular_slider.setValue(int(self.specular_strength * 100))
            self.specular_slider.valueChanged.connect(lambda v: self.update_specular_strength(v / 100.0))
            specular_layout.addWidget(self.specular_slider)
            layout.addLayout(specular_layout)

            # Shininess control
            shininess_layout = QHBoxLayout()
            shininess_label = QLabel("Shininess:")
            shininess_layout.addWidget(shininess_label)
            self.shininess_slider = QSlider()
            self.shininess_slider.setOrientation(1)
            self.shininess_slider.setRange(1, 256)
            self.shininess_slider.setValue(int(self.shininess))
            self.shininess_slider.valueChanged.connect(lambda v: self.update_shininess(float(v)))
            shininess_layout.addWidget(self.shininess_slider)
            layout.addLayout(shininess_layout)

    def _on_color(self, color):
        try:
            self.color = color
            self.flat_rgb = text_to_rgba(color, flat=True)
            self.need_update_setting = True
        except ValueError:
            pass

    def update_wireframe(self, value):
        self.wireframe = value
        
    def update_enable_lighting(self, value):
        self.enable_lighting = value
        self.need_update_setting = True
        
    def update_line_width(self, value):
        self.line_width = value
        self.need_update_setting = True
        
    def update_ambient_strength(self, value):
        self.ambient_strength = value
        self.need_update_setting = True
        
    def update_diffuse_strength(self, value):
        self.diffuse_strength = value
        self.need_update_setting = True
        
    def update_specular_strength(self, value):
        self.specular_strength = value
        self.need_update_setting = True
        
    def update_shininess(self, value):
        self.shininess = value
        self.need_update_setting = True

    def update_alpha(self, value):
        """Update mesh alpha (opacity)"""
        self.alpha = float(value)
        self.need_update_setting = True

    def set_data(self, data):
        """
        Set complete mesh data at once.
        
        Args:
            data: Nx9 numpy array where N is the number of triangles
                  Each row: [v0.x, v0.y, v0.z, v1.x, v1.y, v1.z, v2.x, v2.y, v2.z]
        """
        print("Setting static mesh data with {} triangles".format(len(data)))
        if not isinstance(data, np.ndarray):
            raise ValueError("Data must be a numpy array")

        # Check shape
        if data.ndim != 2 or data.shape[1] != 9:
            # Try to reshape if it's Nx3 (vertex list)
            if data.ndim == 2 and data.shape[1] == 3 and data.shape[0] % 3 == 0:
                data = data.reshape(-1, 9)
            else:
                raise ValueError(f"Invalid data shape {data.shape}. Expected Nx9 or (N*3)x3")

        self.vertices = data.astype(np.float32)
        self.num_triangles = len(self.vertices)
        self.need_update_buffer = True

    def clear_mesh(self):
        """Clear all mesh data"""
        self.vertices = None
        self.num_triangles = 0
        self.need_update_buffer = True

    def initialize_gl(self):
        """OpenGL initialization - load triangle shader"""
        frag_shader = open(self.path + '/../shaders/mesh_frag.glsl', 'r').read()
        
        try:
            vert_shader = open(self.path + '/../shaders/triangle_mesh_vert.glsl', 'r').read()
            geom_shader = open(self.path + '/../shaders/triangle_mesh_geom.glsl', 'r').read()
            self.program = shaders.compileProgram(
                shaders.compileShader(vert_shader, GL_VERTEX_SHADER),
                shaders.compileShader(geom_shader, GL_GEOMETRY_SHADER),
                shaders.compileShader(frag_shader, GL_FRAGMENT_SHADER),
            )
        except Exception as e:
            print(f"Error compiling static mesh shader: {e}")
            raise

    def update_render_buffer(self):
        """
        Update GPU buffer with triangle data.
        Each triangle: 9 floats (3 vertices x 3 coordinates)
        """
        if self.num_triangles == 0 or self.vertices is None:
            return
        
        # Initialize buffers on first call
        if self.vao is None:
            self.vao = glGenVertexArrays(1)
            self.vbo = glGenBuffers(1)
        
        if not self.need_update_buffer:
            return
            
        glBindVertexArray(self.vao)
        glBindBuffer(GL_ARRAY_BUFFER, self.vbo)
        
        # Upload all triangle data
        glBufferData(GL_ARRAY_BUFFER,
                    self.vertices.nbytes,
                    self.vertices,
                    GL_STATIC_DRAW)
        
        # Setup vertex attributes
        # Triangle: [v0.x, v0.y, v0.z, v1.x, v1.y, v1.z, v2.x, v2.y, v2.z]
        # 9 floats = 36 bytes stride
        stride = 36
        
        # v0 (location 1) - vec3
        glEnableVertexAttribArray(1)
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, stride, ctypes.c_void_p(0))
        glVertexAttribDivisor(1, 1)
        
        # v1 (location 2) - vec3
        glEnableVertexAttribArray(2)
        glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, stride, ctypes.c_void_p(12))
        glVertexAttribDivisor(2, 1)
        
        # v2 (location 3) - vec3
        glEnableVertexAttribArray(3)
        glVertexAttribPointer(3, 3, GL_FLOAT, GL_FALSE, stride, ctypes.c_void_p(24))
        glVertexAttribDivisor(3, 1)
        
        glBindVertexArray(0)
        glBindBuffer(GL_ARRAY_BUFFER, 0)
        self.need_update_buffer = False
        
    def update_setting(self):
        """Set rendering parameters"""
        if not self.need_update_setting:
            return
        
        set_uniform(self.program, int(self.enable_lighting), 'if_light')
        set_uniform(self.program, 1, 'two_sided')
        set_uniform(self.program, np.array(self.light_color, dtype=np.float32), 'light_color')
        set_uniform(self.program, float(self.ambient_strength), 'ambient_strength')
        set_uniform(self.program, float(self.diffuse_strength), 'diffuse_strength')
        set_uniform(self.program, float(self.specular_strength), 'specular_strength')
        set_uniform(self.program, float(self.shininess), 'shininess')
        set_uniform(self.program, float(self.alpha), 'alpha')
        set_uniform(self.program, int(self.flat_rgb), 'flat_rgb')
        self.need_update_setting = False

    def paint(self):
        """
        Render the static mesh using instanced rendering with geometry shader.
        Each triangle instance is rendered as a point, geometry shader generates 1 triangle.
        """
        if self.num_triangles == 0 or self.vertices is None:
            return
        
        glUseProgram(self.program)
    
        self.update_render_buffer()
        self.update_setting()
        
        view_matrix = self.glwidget().view_matrix
        set_uniform(self.program, view_matrix, 'view')
        project_matrix = self.glwidget().projection_matrix
        set_uniform(self.program, project_matrix, 'projection')
        view_pos = self.glwidget().center
        set_uniform(self.program, np.array(view_pos), 'view_pos')
            
        # Enable blending and depth testing
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glEnable(GL_DEPTH_TEST)
        glDisable(GL_CULL_FACE)  # two-sided rendering
        
        # Set line width
        glLineWidth(self.line_width)
        
        # Bind VAO
        glBindVertexArray(self.vao)
                
        if self.wireframe:
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        else:
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        
        # Draw using instanced rendering
        # Input: POINTS (one per triangle instance)
        # Geometry shader generates 1 triangle (3 vertices) per point
        glDrawArraysInstanced(GL_POINTS, 0, 1, self.num_triangles)
            
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        glBindVertexArray(0)
        glDisable(GL_DEPTH_TEST)
        glDisable(GL_BLEND)
        glUseProgram(0)



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""


from q3dviewer.base_item import BaseItem
from OpenGL.GL import *
import numpy as np

try:
    from OpenGL.GLUT import glutBitmapCharacter, glutInit
    from OpenGL.GLUT import (
        GLUT_BITMAP_HELVETICA_10,
        GLUT_BITMAP_HELVETICA_12,
        GLUT_BITMAP_HELVETICA_18,
        GLUT_BITMAP_TIMES_ROMAN_24,
    )
    GLUT_AVAILABLE = True
except ImportError:
    GLUT_AVAILABLE = False
    print("Warning: GLUT not available. Text will not be rendered.")

def get_glut_font(font_size):
    """Get GLUT font based on size, only if GLUT is available"""
    if not GLUT_AVAILABLE:
        return None
    
    # Map requested font_size to a GLUT font object
    if font_size <= 10:
        return GLUT_BITMAP_HELVETICA_10
    elif font_size <= 12:
        return GLUT_BITMAP_HELVETICA_12
    elif font_size <= 18:
        return GLUT_BITMAP_HELVETICA_18
    elif font_size <= 24:
        return GLUT_BITMAP_TIMES_ROMAN_24
    else:
        return GLUT_BITMAP_TIMES_ROMAN_24  # Largest available
    

# draw points with color (x, y, z, color)
class Text3DItem(BaseItem):
    """
    A OpenGL 3d text and mark item.

    Attributes:
        data: list storing text data. Each element is a dict with keys:
            'text': str, the text to display
            'position': (x, y, z), the 3D position of the text
            'color': (r, g, b, a), the color of the text
            'font_size': float, the font size of the text
            'point_size': float, size of point to draw at position (0 for no point)
            'line_width': float, width of line to draw between points (0 for no line)

    """
    def __init__(self, data=[]):
        super().__init__()
        self._disable_setting = True
        self.data_list = data # map of {'text': str, 'position': (x,y,z), 'color': (r,g,b,a), 'size': float}

    def add_setting(self, layout):
        pass # No settings for Text3DItem

    def set_data(self, data, append=False):
        if not append:
            self.data_list = []
        self.data_list.extend(data)

    def clear_data(self):
        self.data_list = []


    def initialize_gl(self):
        """Initialize OpenGL resources, with GLUT fallback handling"""
        global GLUT_AVAILABLE
        if GLUT_AVAILABLE:
            try:
                # Check if glutInit is actually callable before calling it
                if hasattr(glutInit, '__call__') and bool(glutInit):
                    glutInit()
                else:
                    print("Warning: glutInit not callable, using fallback text rendering")
                    GLUT_AVAILABLE = False
            except Exception as e:
                print(f"Warning: GLUT initialization failed: {e}. Using fallback text rendering.")
                GLUT_AVAILABLE = False
        # super().initialize_gl()


    def paint(self):
        for item in self.data_list:
            # Handle both dictionary and string formats
            if isinstance(item, dict):
                text = item.get('text', '')
                pos = item.get('position', (0.0, 0.0, 0.0))
                font_size = item.get('font_size', 24)
                color = item.get('color', (1.0, 1.0, 1.0, 1.0))
                point_size = item.get('point_size', 0.0)
            elif isinstance(item, str):
                # If item is a string, treat it as text with default position and color
                text = item
                pos = (0.0, 0.0, 0.0)
                color = (1.0, 1.0, 1.0, 1.0)
                font_size = 24
                point_size = 0.0
            else:
                print(f"Warning: Unsupported item type: {type(item)}")
                continue

            # Convert numpy array to tuple if needed
            if isinstance(pos, np.ndarray):
                pos = tuple(pos.astype(float))
            
            glColor4f(*color)
            
            if point_size > 0.0:
                # draw a point at the position
                glPointSize(point_size)
                glBegin(GL_POINTS)
                glVertex3f(*pos)
                glEnd()
            
            # Text rendering with GLUT fallback
            if GLUT_AVAILABLE:
                offset = 0.02
                pos_text = (pos[0] + offset, pos[1] + offset, pos[2] + offset)
                glRasterPos3f(*pos_text)
                font = get_glut_font(font_size)
                if font is not None:
                    try:
                        for ch in text:
                            glutBitmapCharacter(font, ord(ch))
                    except Exception as e:
                        print(f"Error rendering text '{text}': {e}")

        # draw lines between points
        for i in range(len(self.data_list) - 1):
            item1 = self.data_list[i]
            item2 = self.data_list[i + 1]
            if isinstance(item1, dict) and isinstance(item2, dict):
                pos1 = item1.get('position', (0.0, 0.0, 0.0))
                pos2 = item2.get('position', (0.0, 0.0, 0.0))
                line_width = item1.get('line_width', 0.0)
                color = item1.get('color', (1.0, 1.0, 1.0, 1.0))
                if line_width > 0.0:
                    glColor4f(*color)
                    glLineWidth(line_width)
                    glBegin(GL_LINES)
                    glVertex3f(*pos1)
                    glVertex3f(*pos2)
                    glEnd()


#!/usr/bin/env python3

"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""


import numpy as np
from q3dviewer.base_item import BaseItem
from OpenGL.GL import *
from OpenGL.GL import shaders
from q3dviewer.Qt.QtWidgets import QLabel, QCheckBox, QDoubleSpinBox, QSlider, QHBoxLayout, QLineEdit

import os
from q3dviewer.utils import set_uniform, text_to_rgba
import time


class MeshItem(BaseItem):
    """
    A OpenGL mesh item for rendering 3D triangular meshes.
    Attributes:
        color (str or tuple): Accepts any valid matplotlib color (e.g., 'red', '#FF4500', (1.0, 0.5, 0.0)).
        wireframe (bool): If True, renders the mesh in wireframe mode.
    """
    # Class-level constants
    FACE_CAPACITY = 1000000    # Initial capacity for faces
    BIG_INT = 2**31 - 1         # Sentinel value for dirty region tracking
    FACE_INPUT_DTYPE = np.dtype([
        ('key', np.int64),
        ('vertices', np.float32, (12,)),
        ('good', np.uint8)
    ])
    
    def __init__(self, color='lightblue', wireframe=False):
        super(MeshItem, self).__init__()
        self.wireframe = wireframe
        self.color = color
        self.flat_rgb = text_to_rgba(color, flat=True)
        
        # Faces buffer: N x 13 numpy array
        # Each row: [v0.x, v0.y, v0.z, v1.x, v1.y, v1.z, v2.x, v2.y, v2.z, v3.x, v3.y, v3.z, good]
        self.faces = np.zeros((self.FACE_CAPACITY, 13), dtype=np.float32)
        
        # valid_f_top: pointer to end of valid faces
        self.valid_f_top = 0
        
        # Dirty region tracking for efficient GPU updates
        # dirty_min: start index of modified region (inclusive)
        # dirty_max: end index of modified region (exclusive)
        self.dirty_min = self.BIG_INT
        self.dirty_max = 0
        
        # key2index: mapping from face_key to face buffer index
        self.key2index = {}  # {face_key: face_index}
        
        # OpenGL objects
        self.vao = None
        self.vbo = None
        self.program = None
        self._gpu_face_capacity = 0    # Track GPU buffer capacity
        
        # Fixed rendering parameters (not adjustable via UI)
        self.enable_lighting = True
        self.line_width = 1.0
        self.light_pos = [1.0, 1.0, 1.0]
        self.light_color = [1.0, 1.0, 1.0]
        self.ambient_strength = 0.1
        self.diffuse_strength = 1.2
        self.specular_strength = 0.1
        self.shininess = 32.0
        self.alpha = 1.0
        
        # Settings flag
        self.need_update_setting = True
        self.need_update_buffer = True
        self.path = os.path.dirname(__file__)
    
        
    def add_setting(self, layout):
        """Add UI controls for mesh visualization"""
        # Only keep wireframe toggle - all other parameters are fixed
        self.wireframe_box = QCheckBox("Wireframe Mode")
        self.wireframe_box.setChecked(self.wireframe)
        self.wireframe_box.toggled.connect(self.update_wireframe)
        layout.addWidget(self.wireframe_box)

        # Enable lighting toggle
        self.lighting_box = QCheckBox("Enable Lighting")
        self.lighting_box.setChecked(self.enable_lighting)
        self.lighting_box.toggled.connect(self.update_enable_lighting)
        layout.addWidget(self.lighting_box)

        label_rgb = QLabel("Color:")
        label_rgb.setToolTip("Use hex color, i.e. #FF4500, or named color, i.e. 'red'")
        layout.addWidget(label_rgb)
        self.edit_rgb = QLineEdit()
        self.edit_rgb.setToolTip("Use hex color, i.e. #FF4500, or named color, i.e. 'red'")
        self.edit_rgb.setText(self.color)
        self.edit_rgb.textChanged.connect(self._on_color)
        layout.addWidget(self.edit_rgb)

        # Material property controls for Phong lighting
        if self.enable_lighting:
            # Ambient strength control (slider 0-100 mapped to 0.0-1.0)
            ambient_layout = QHBoxLayout()
            ambient_label = QLabel("Ambient Strength:")
            ambient_layout.addWidget(ambient_label)
            self.ambient_slider = QSlider()
            self.ambient_slider.setOrientation(1)  # Qt.Horizontal
            self.ambient_slider.setRange(0, 100)
            self.ambient_slider.setValue(int(self.ambient_strength * 100))
            self.ambient_slider.valueChanged.connect(lambda v: self.update_ambient_strength(v / 100.0))
            ambient_layout.addWidget(self.ambient_slider)
            layout.addLayout(ambient_layout)

            # Diffuse strength control (slider 0-200 mapped to 0.0-2.0)
            diffuse_layout = QHBoxLayout()
            diffuse_label = QLabel("Diffuse Strength:")
            diffuse_layout.addWidget(diffuse_label)
            self.diffuse_slider = QSlider()
            self.diffuse_slider.setOrientation(1)
            self.diffuse_slider.setRange(0, 200)
            self.diffuse_slider.setValue(int(self.diffuse_strength * 100))
            self.diffuse_slider.valueChanged.connect(lambda v: self.update_diffuse_strength(v / 100.0))
            diffuse_layout.addWidget(self.diffuse_slider)
            layout.addLayout(diffuse_layout)

            # Specular strength control (slider 0-200 mapped to 0.0-2.0)
            specular_layout = QHBoxLayout()
            specular_label = QLabel("Specular Strength:")
            specular_layout.addWidget(specular_label)
            self.specular_slider = QSlider()
            self.specular_slider.setOrientation(1)
            self.specular_slider.setRange(0, 200)
            self.specular_slider.setValue(int(self.specular_strength * 100))
            self.specular_slider.valueChanged.connect(lambda v: self.update_specular_strength(v / 100.0))
            specular_layout.addWidget(self.specular_slider)
            layout.addLayout(specular_layout)

            # Shininess control (slider 1-256 mapped to 1-256)
            shininess_layout = QHBoxLayout()
            shininess_label = QLabel("Shininess:")
            shininess_layout.addWidget(shininess_label)
            self.shininess_slider = QSlider()
            self.shininess_slider.setOrientation(1)
            self.shininess_slider.setRange(1, 256)
            self.shininess_slider.setValue(int(self.shininess))
            self.shininess_slider.valueChanged.connect(lambda v: self.update_shininess(float(v)))
            shininess_layout.addWidget(self.shininess_slider)
            layout.addLayout(shininess_layout)

    def _on_color(self, color):
        try:
            self.color = color
            self.flat_rgb = text_to_rgba(color, flat=True)
            self.need_update_setting = True
        except ValueError:
            pass

    def update_wireframe(self, value):
        self.wireframe = value
        
    def update_enable_lighting(self, value):
        self.enable_lighting = value
        self.need_update_setting = True
        
    def update_line_width(self, value):
        self.line_width = value
        self.need_update_setting = True
        
    def update_ambient_strength(self, value):
        self.ambient_strength = value
        self.need_update_setting = True
        
    def update_diffuse_strength(self, value):
        self.diffuse_strength = value
        self.need_update_setting = True
        
    def update_specular_strength(self, value):
        self.specular_strength = value
        self.need_update_setting = True
        
    def update_shininess(self, value):
        self.shininess = value
        self.need_update_setting = True

    def update_alpha(self, value):
        """Update mesh alpha (opacity)"""
        self.alpha = float(value)
        self.need_update_setting = True

    def set_data(self, data):
        """
        Args:
            data: One of the following formats:
                  - Nx3 numpy array (N must be divisible by 3): vertex list -> static
                  - Nx9 numpy array: triangle list -> static
                  - Structured array with dtype [('key', int64), ('vertices', float32, (12,)), ('good', uint32)] -> incremental
        """
        if not isinstance(data, np.ndarray):
            raise ValueError("Data must be a numpy array")
        
        # Structured array format -> use incremental path (has keys for updates)
        if data.dtype == self.FACE_INPUT_DTYPE:
            self.set_incremental_data(data)
            return
        
        # Nx3 or Nx9 format -> use static path (more efficient, no key overhead)
        if data.ndim == 2 and data.shape[1] in [3, 9]:
            self.set_static_data(data)
            return
        
        raise ValueError("Data must be Nx3, Nx9, or structured array format")

    def set_static_data(self, data):
        """
        Efficiently set static mesh data without key2index overhead.
        For static meshes that don't need incremental updates.
        
        Args:
            data: numpy array in one of these formats:
                  - Nx3: vertex list (N must be divisible by 3)
                  - Nx9: triangle list
        """
        if not isinstance(data, np.ndarray):
            raise ValueError("Data must be a numpy array")
        
        if data.ndim != 2:
            raise ValueError(f"Data must be 2D array, got {data.ndim}D")
        self.clear_mesh()
        # Handle Nx3 format
        if data.shape[1] == 3:
            if data.shape[0] % 3 != 0:
                raise ValueError(f"Nx3 format requires N divisible by 3, got N={data.shape[0]}")
            
            num_faces = data.shape[0] // 3
            faces = np.zeros((num_faces, 13), dtype=np.float32)
            
            # Reshape to (num_faces, 9) for efficient copying
            tmp = data.reshape(num_faces, 9)
            faces[:, 0:3]  = tmp[:, 0:3]   # v0
            faces[:, 3:6]  = tmp[:, 3:6]   # v1
            faces[:, 6:9]  = tmp[:, 6:9]   # v2
            faces[:, 9:12] = tmp[:, 6:9]   # v3 = v2 (degenerate)
            faces[:, 12] = 1.0             # good=1.0
            
        # Handle Nx9 format
        elif data.shape[1] == 9:
            num_faces = data.shape[0]
            faces = np.zeros((num_faces, 13), dtype=np.float32)
            
            faces[:, 0:9] = data           # Copy all 9 vertices
            faces[:, 9:12] = data[:, 6:9]  # v3 = v2 (degenerate)
            faces[:, 12] = 1.0             # good=1.0
            
        else:
            raise ValueError(f"Data shape must be Nx3 or Nx9, got Nx{data.shape[1]}")
        
        # Replace faces buffer (static data, no key management)
        self.clear_mesh()
        self.faces = faces
        self.valid_f_top = num_faces
        
        # Mark entire buffer as dirty
        self.dirty_min = 0
        self.dirty_max = self.valid_f_top
        self.need_update_buffer = True


    def set_incremental_data(self, fs):
        """
        Incrementally update mesh with new face data.
        Args:
            fs: Structured numpy array with dtype:
                [('key', np.int64), ('vertices', np.float32, (12,)), ('good', np.uint32)]
                - key: unique identifier for the face
                - vertices: 12 floats representing 4 vertices (v0, v1, v2, v3)
                - good: 0 or 1, whether to render this face
        Updates:
            - faces: updates existing faces or appends new ones
            - key2index: tracks face_key -> face_index mapping
        """
        if fs is None or len(fs) == 0:
            return
        
        if not isinstance(fs, np.ndarray) or fs.dtype.names is None:
            raise ValueError("fs must be a structured numpy array with fields: key, vertices, good")

        # Ensure enough capacity in faces buffer
        while self.valid_f_top + len(fs) > len(self.faces):
            self._expand_face_buffer()

        # Prepare face data: convert structured array to Nx13 format
        n_faces = len(fs)
        face_data = np.zeros((n_faces, 13), dtype=np.float32)
        
        # Copy vertices (12 floats -> positions 0:12)
        face_data[:, :12] = fs['vertices']
        
        # Copy good flag (position 12)
        face_data[:, 12] = fs['good'].astype(np.float32)
        
        # Extract keys
        keys = fs['key']
        
        # Optimization: Separate updates from new insertions
        update_mask = np.array([key in self.key2index for key in keys], dtype=bool)
        new_mask = ~update_mask
        
        # Batch update existing faces
        if np.any(update_mask):
            update_keys = keys[update_mask]
            update_indices = np.array([self.key2index[key] for key in update_keys], dtype=np.int32)
            self.faces[update_indices] = face_data[update_mask]
            
            # Update dirty region for modified faces
            self.dirty_min = min(self.dirty_min, int(np.min(update_indices)))
            self.dirty_max = max(self.dirty_max, int(np.max(update_indices) + 1))
            self.need_update_buffer = True
        
        # Batch insert new faces
        if np.any(new_mask):
            new_keys = keys[new_mask]
            new_face_data = face_data[new_mask]
            n_new = len(new_keys)
            
            # Update dirty region for new faces
            start_index = self.valid_f_top            
            # Insert data
            self.faces[start_index: start_index + n_new] = new_face_data
            
            # Update key2index mapping for new faces
            for i, face_key in enumerate(new_keys):
                self.key2index[face_key] = start_index + i
            self.valid_f_top += n_new
            self.dirty_min = min(self.dirty_min, start_index)
            self.dirty_max = max(self.dirty_max, start_index + n_new)
            self.need_update_buffer = True
    
    def _expand_face_buffer(self):
        """Expand the faces buffer when capacity is reached"""
        new_capacity = len(self.faces) + self.FACE_CAPACITY
        new_buffer = np.zeros((new_capacity, 13), dtype=np.float32)
        new_buffer[:len(self.faces)] = self.faces
        self.faces = new_buffer
    
    def clear_mesh(self):
        """Clear all mesh data and reset buffers"""
        self.valid_f_top = 0
        self.dirty_min = self.BIG_INT
        self.dirty_max = 0
        self.key2index.clear()
        if hasattr(self, 'indices_array'):
            self.indices_array = np.array([], dtype=np.uint32)

    def initialize_gl(self):
        """OpenGL initialization"""
        # Use instanced mesh shaders with geometry shader for GPU-side triangle generation
        vert_shader = open(self.path + '/../shaders/mesh_vert.glsl', 'r').read()
        geom_shader = open(self.path + '/../shaders/mesh_geom.glsl', 'r').read()
        frag_shader = open(self.path + '/../shaders/mesh_frag.glsl', 'r').read()
        try:
            program = shaders.compileProgram(
                shaders.compileShader(vert_shader, GL_VERTEX_SHADER),
                shaders.compileShader(geom_shader, GL_GEOMETRY_SHADER),
                shaders.compileShader(frag_shader, GL_FRAGMENT_SHADER),
            )
            self.program = program
        except Exception as e:
            raise

    def update_render_buffer(self):
        """
        Update GPU buffer with face data (no separate vertex buffer).
        Each face contains embedded vertex positions (13 floats).
        Geometry shader generates triangles on GPU from face vertices.
        Dynamically resizes GPU buffer when Python buffer expands.
        """
        if self.valid_f_top == 0:
            return
        
        # Initialize buffers on first call
        if self.vao is None:
            self.vao = glGenVertexArrays(1)
            self.vbo = glGenBuffers(1)
            self._gpu_face_capacity = 0
        
        # Check if we need to reallocate VBO for faces
        if self._gpu_face_capacity < len(self.faces):
            glBindVertexArray(self.vao)
            glBindBuffer(GL_ARRAY_BUFFER, self.vbo)
            glBufferData(GL_ARRAY_BUFFER,
                        self.faces.nbytes,
                        None,
                        GL_DYNAMIC_DRAW)
            
            # Setup face attributes (per-instance)
            # Face data: [v0.x, v0.y, v0.z, v1.x, v1.y, v1.z, v2.x, v2.y, v2.z, v3.x, v3.y, v3.z, good]
            # 13 floats = 52 bytes stride
            
            # v0 (location 1) - vec3
            glEnableVertexAttribArray(1)
            glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 52, ctypes.c_void_p(0))
            glVertexAttribDivisor(1, 1)
            
            # v1 (location 2) - vec3
            glEnableVertexAttribArray(2)
            glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 52, ctypes.c_void_p(12))
            glVertexAttribDivisor(2, 1)
            
            # v2 (location 3) - vec3
            glEnableVertexAttribArray(3)
            glVertexAttribPointer(3, 3, GL_FLOAT, GL_FALSE, 52, ctypes.c_void_p(24))
            glVertexAttribDivisor(3, 1)
            
            # v3 (location 4) - vec3
            glEnableVertexAttribArray(4)
            glVertexAttribPointer(4, 3, GL_FLOAT, GL_FALSE, 52, ctypes.c_void_p(36))
            glVertexAttribDivisor(4, 1)
            
            # good flag (location 5) - float
            glEnableVertexAttribArray(5)
            glVertexAttribPointer(5, 1, GL_FLOAT, GL_FALSE, 52, ctypes.c_void_p(48))
            glVertexAttribDivisor(5, 1)
            
            glBindVertexArray(0)
            glBindBuffer(GL_ARRAY_BUFFER, 0)
            self._gpu_face_capacity = len(self.faces)
        
        # Upload faces to VBO (only dirty region)
        if self.need_update_buffer:
            glBindBuffer(GL_ARRAY_BUFFER, self.vbo)
            
            # Calculate the range to upload [dirty_min, dirty_max)
            start_index = int(self.dirty_min)
            end_index = int(self.dirty_max)
            count = end_index - start_index
            
            # Upload only the modified region
            glBufferSubData(GL_ARRAY_BUFFER,
                           start_index * 13 * 4,  # offset in bytes
                           count * 13 * 4,         # size in bytes
                           self.faces[start_index:end_index])
            
            glBindBuffer(GL_ARRAY_BUFFER, 0)
            self.need_update_buffer = False
            
            # Reset dirty region
            self.dirty_min = self.BIG_INT
            self.dirty_max = 0
        
    def update_setting(self):
        """Set fixed rendering parameters (called once during initialization)"""
        if not self.need_update_setting:
            return
        # Set fixed uniforms for instanced shaders
        set_uniform(self.program, int(self.enable_lighting), 'if_light')
        set_uniform(self.program, 1, 'two_sided')

        set_uniform(self.program, np.array(self.light_color, dtype=np.float32), 'light_color')
        set_uniform(self.program, float(self.ambient_strength), 'ambient_strength')
        set_uniform(self.program, float(self.diffuse_strength), 'diffuse_strength')
        set_uniform(self.program, float(self.specular_strength), 'specular_strength')
        set_uniform(self.program, float(self.shininess), 'shininess')
        set_uniform(self.program, float(self.alpha), 'alpha')
        set_uniform(self.program, int(self.flat_rgb), 'flat_rgb')
        self.need_update_setting = False

    def paint(self):
        """
        Render the mesh using instanced rendering with geometry shader.
        Each face instance is rendered as a point, geometry shader generates 2 triangles.
        GPU filters faces based on good flag.
        """
        if self.valid_f_top == 0:
            return
        
        glUseProgram(self.program)
    
        self.update_render_buffer()
        self.update_setting()
        
        view_matrix = self.glwidget().view_matrix
        set_uniform(self.program, view_matrix, 'view')
        project_matrix = self.glwidget().projection_matrix
        set_uniform(self.program, project_matrix, 'projection')
        view_pos = self.glwidget().center
        set_uniform(self.program, np.array(view_pos), 'view_pos')
            
        # Enable blending and depth testing
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glEnable(GL_DEPTH_TEST)
        glDisable(GL_CULL_FACE)  # two-sided rendering
        
        # Set line width
        glLineWidth(self.line_width)
        
        # Bind VAO (vertex positions are now in VBO attributes)
        glBindVertexArray(self.vao)
                
        if self.wireframe:
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        else:
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        
        # Draw using instanced rendering
        # Input: POINTS (one per face instance)
        # Geometry shader generates 2 triangles (6 vertices) per point
        glDrawArraysInstanced(GL_POINTS, 0, 1, self.valid_f_top)
            
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        glBindVertexArray(0)
        glDisable(GL_DEPTH_TEST)
        glDisable(GL_BLEND)
        glUseProgram(0)
        
